"""
Module for Switchboards.

Note that additional args and kwargs for train or execute are currently not
supported.
"""
from __future__ import division
from builtins import range
from builtins import object
from past.utils import old_div


import mdp
from mdp import numx


class SwitchboardException(mdp.NodeException):
    """Exception for routing problems in the Switchboard class."""
    pass


# TODO: deal with input_dim, output_dim and dtype correctly,
#    like in IdentityNode

class Switchboard(mdp.Node):
    """Does the routing associated with the connections between layers.

    It may be directly used as a layer/node, routing all the data at once. If
    the routing/mapping is not injective the processed data may be quite large
    and probably contains many redundant copies of the input data.
    So is this case one may instead use nodes for individual output
    channels and put each in a MultiNode.

    SwitchboardLayer is the most general version of a switchboard layer, since
    there is no imposed rule for the connection topology. For practical
    applications should often derive more specialized classes.
    """

    def __init__(self, input_dim, connections):
        """Create a generic switchboard.

        The input and output dimension as well as dtype have to be fixed
        at initialization time.

        Keyword arguments:
        input_dim -- Dimension of the input data (number of connections).
        connections -- 1d Array or sequence with an entry for each output
            connection, containing the corresponding index of the
            input connection.
        """
        # check connections for inconsistencies
        if len(connections) == 0:
            err = "Received empty connection list."
            raise SwitchboardException(err)
        if numx.nanmax(connections) >= input_dim:
            err = ("One or more switchboard connection "
                   "indices exceed the input dimension.")
            raise SwitchboardException(err)
        # checks passed
        self.connections = numx.array(connections)
        output_dim = len(connections)
        super(Switchboard, self).__init__(input_dim=input_dim,
                                          output_dim=output_dim)
        # try to invert connections
        if (self.input_dim == self.output_dim and
            len(numx.unique(self.connections)) == self.input_dim):
            self.inverse_connections = numx.argsort(self.connections)
        else:
            self.inverse_connections = None

    def _execute(self, x):
        return x[:, self.connections]

    @staticmethod
    def is_trainable():
        return False

    def is_invertible(self):
        if self.inverse_connections is None:
            return False
        else:
            return True

    def _inverse(self, x):
        if self.inverse_connections is None:
            raise SwitchboardException("Connections are not invertible.")
        else:
            return x[:, self.inverse_connections]

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('AllFloat') +
                mdp.utils.get_dtypes('AllInteger') +
                mdp.utils.get_dtypes('Character'))

class MeanInverseSwitchboard(Switchboard):
    """Variant of Switchboard with modified inverse.

    If the switchboard mapping is not injective, then the mean
    values are used for the inverse. Inputs that are discarded in the
    mapping are set to zero.

    You can use this class as a mixin for other switchboard classes.
    """

    def _inverse(self, x):
        """Take the mean of overlapping values."""
        n_y_cons = numx.bincount(self.connections)  # n. connections to y_i
        y_cons = numx.argsort(self.connections)  # x indices for y_i
        y = numx.zeros((len(x), self.input_dim))
        i_x_counter = 0  # counter for processed x indices
        i_y = 0  # current y index
        while True:
            n_cons = n_y_cons[i_y]
            if n_cons > 0:
                y[:,i_y] = old_div(numx.sum(x[:,y_cons[i_x_counter:
                                               i_x_counter + n_cons]],
                                    axis=1), n_cons)
                i_x_counter += n_cons
                if i_x_counter >= self.output_dim:
                    break
            i_y += 1
        return y

    @staticmethod
    def is_invertible():
        return True


class ChannelSwitchboard(Switchboard):
    """Base class for Switchboards in which the data is bundled into channels.

    The dimensions of the input / output channels are constant.

    public attributes (in addition to inherited attributes):
        out_channel_dim
        in_channel_dim
        output_channels
    """

    def __init__(self, input_dim, connections, out_channel_dim,
                 in_channel_dim=1):
        """Initialize the switchboard.

        connections -- Connection sequence like for a standard switchboard
            (the indices do not correspond to whole channels, but single
             connections).
        out_channel_dim -- Number of connections per output channel.
        in_channel_dim -- Number of connections per input channel (default 1).
            All the components of an input channel are treated equally
            by the switchboard (i.e., they are routed to the same output
            channel).
        """
        super(ChannelSwitchboard, self).__init__(input_dim, connections)
        # perform checks
        if self.output_dim % out_channel_dim:
            err = ("Output dim %d is not multiple of out_channel_dim %d." %
                   (self.output_dim, out_channel_dim))
            raise SwitchboardException(err)
        if input_dim % in_channel_dim:
            err = ("Input dim %d is not multiple of in_channel_dim %d." %
                   (self.input_dim, in_channel_dim))
            raise SwitchboardException(err)
        # finalize initialization
        self.out_channel_dim = out_channel_dim
        self.in_channel_dim = in_channel_dim
        self.output_channels = self.output_dim // out_channel_dim
        self.input_channels = self.input_dim // in_channel_dim

    def get_out_channel_input(self, channel):
        """Return the input connections for the given channel index.

        channel -- index of the requested channel (starting at 0)
        """
        index = channel * self.out_channel_dim
        return self.connections[index : index+self.out_channel_dim]

    def get_out_channel_node(self, channel):
        """Return a Switchboard that does the routing for a single
        output channel.

        channel -- index of the requested channel (starting at 0)
        """
        return Switchboard(self.input_dim, self.get_out_channel_input(channel))

    def get_out_channels_input_channels(self, channels):
        """Return array of input channel indices for the given output channels.

        channels -- Sequence of the requested output channels or a single
            channel index (i.e. a number).

        The returned array contains the indices of all input channels which
        are connected to at least one of the given output channels.
        """
        if isinstance(channels, int):
            channels = [channels]
        # create boolean arry to determine with active inputs
        channels_input = self.connections.reshape((-1, self.out_channel_dim))
        channels_input = channels_input[channels].reshape(-1)
        covered = numx.zeros(self.input_dim, dtype="bool")
        covered[channels_input] = True
        # reshape to perform logical OR over the input channels
        covered = covered.reshape((-1, self.in_channel_dim))
        covered = covered.sum(axis=1, dtype=bool)
        return covered.nonzero()[0]


def to_2tuple(value):
    """Return value or (value, value) if value is not a tuple."""
    if isinstance(value, tuple):
        return value
    if isinstance(value, list):
        return tuple(value)
    return (value, value)

    
class Rectangular2dSwitchboardException(SwitchboardException):
    """Exception for routing problems in the Rectangular2dSwitchboard class."""
    pass


class Rectangular2dSwitchboard(ChannelSwitchboard):
    """Switchboard for a 2-dimensional topology.

    This is a specialized version of SwitchboardLayer that makes it easy to
    implement connection topologies which are based on a 2-dimensional network
    layers.

    The input connections are assumed to be grouped into so called channels,
    which are considered as lying in a two dimensional rectangular plane.
    Each output channel corresponds to a 2d rectangular field in the
    input plane. The fields can overlap.

    The coordinates follow the standard image convention (see the above
    CoordinateTranslator class).

    public attributes (in addition to init arguments and inherited attributes):
        unused_channels_xy
        out_channels_xy
    """
    
    def __init__(self, in_channels_xy, field_channels_xy,
                 field_spacing_xy=1, in_channel_dim=1, ignore_cover=False):
        """Calculate the connections.

        Keyword arguments:
        in_channels_xy -- 2-Tuple with number of input channels in the x- and
            y-direction (or a single number for both). This has to be
            specified, since the actual input is only one 1d array.
        field_channels_xy -- 2-Tuple with number of channels in each field in
            the x- and y-direction (or a single number for both).
        field_spacing_xy -- 2-Tuple with offset between two fields in the x-
            and y-direction (or a single number for both).
        in_channel_dim -- Number of connections per input channel.
        ignore_cover -- Boolean value defines if an
            Rectangular2dSwitchboardException is raised when the fields do not
            cover all input channels. Set this to True if you are willing to
            risk loosing input channels at the border.
        """
        in_channels_xy = to_2tuple(in_channels_xy)
        field_channels_xy = to_2tuple(field_channels_xy)
        field_spacing_xy = to_2tuple(field_spacing_xy)
        self.in_channels_xy = in_channels_xy
        self.field_channels_xy = field_channels_xy
        self.field_spacing_xy = field_spacing_xy
        # number of channels which are not covered
        out_channel_dim = (in_channel_dim *
                           field_channels_xy[0] * field_channels_xy[1])
        ## check parameters for inconsistencies
        for i, name in enumerate(["x", "y"]):
            if (field_channels_xy[i] > in_channels_xy[i]):
                err = ("Number of field channels exceeds the number of "
                       "input channels in %s-direction. "
                       "This would lead to an empty connection list." %
                       name)
                raise Rectangular2dSwitchboardException(err)
        # number of output channels in x-direction
        x_out_channels = (((in_channels_xy[0] - field_channels_xy[0]) //
                           field_spacing_xy[0]) + 1)
        x_unused_channels = in_channels_xy[0] - field_channels_xy[0]
        if x_unused_channels > 0:
            x_unused_channels %= field_spacing_xy[0]
        elif x_unused_channels < 0:
            x_unused_channels = in_channels_xy[0]
        # number of output channels in y-direction
        y_out_channels = (((in_channels_xy[1] - field_channels_xy[1]) //
                           field_spacing_xy[1]) + 1)
        y_unused_channels = in_channels_xy[1] - field_channels_xy[1]
        if y_unused_channels > 0:
            y_unused_channels %= field_spacing_xy[1]
        elif y_unused_channels < 0:
            y_unused_channels = in_channels_xy[1]
        self.unused_channels_xy = (x_unused_channels, y_unused_channels)
        for i, name in enumerate(["x", "y"]):
            if self.unused_channels_xy[i] and not ignore_cover:
                err = ("Channel fields do not "
                       "cover all input channels in %s-direction." % name)
                raise Rectangular2dSwitchboardException(err)
        self.out_channels_xy = (x_out_channels, y_out_channels)
        ## end of parameters checks
        # TODO: rewrite all this leveraging numpy?
        #    this might make the code much simpler
        #    use 3rd shape entry for channel width
        out_channels = x_out_channels * y_out_channels
        in_trans = CoordinateTranslator(*in_channels_xy)
        # input-output mapping of connections
        # connections has an entry for each output connection,
        # containing the index of the input connection.
        connections = numx.zeros([out_channels * out_channel_dim],
                                 dtype=numx.int32)
        first_out_con = 0
        for y_out_chan in range(self.out_channels_xy[1]):
            for x_out_chan in range(self.out_channels_xy[0]):
                # inner loop over field
                x_start_chan = x_out_chan * field_spacing_xy[0]
                y_start_chan = y_out_chan * field_spacing_xy[1]
                for y_in_chan in range(y_start_chan,
                                       y_start_chan + field_channels_xy[1]):
                    for x_in_chan in range(x_start_chan,
                                        x_start_chan + field_channels_xy[0]):
                        first_in_con = (in_trans.image_to_index(
                                                    x_in_chan, y_in_chan) *
                                        in_channel_dim)
                        connections[first_out_con:
                                    first_out_con + in_channel_dim] = \
                            list(range(first_in_con, first_in_con + in_channel_dim))
                        first_out_con += in_channel_dim
        super(Rectangular2dSwitchboard, self).__init__(
                                input_dim=(in_channel_dim *
                                    in_channels_xy[0] * in_channels_xy[1]),
                                connections=connections,
                                out_channel_dim=out_channel_dim,
                                in_channel_dim=in_channel_dim)


class DoubleRect2dSwitchboardException(SwitchboardException):
    """Exception for routing problems in the DoubleRect2dSwitchboard class."""
    pass


class DoubleRect2dSwitchboard(ChannelSwitchboard):
    """Special 2d Switchboard where each inner point is covered twice.

    First the input is covered with non-overlapping rectangular fields.
    Then the input is covered with fields of the same size that are shifted
    in the x and y direction by half the field size (we call this the
    uneven fields).

    Note that the output of this switchboard cannot be interpreted as
    a rectangular grid, because the short rows are shifted. Instead it is
    a rhombic grid (it is not a hexagonal grid because the distances of the
    field centers do not satisfy the necessary relation).
    See http://en.wikipedia.org/wiki/Lattice_(group)

    Example for a 6x4 input and a field size of 2 in both directions:

    long row fields:

    1 1 2 2 3 3
    1 1 2 2 3 3
    4 4 5 5 6 6
    4 4 5 5 6 6

    short row fields:

    * * * * * *
    * 7 7 8 8 *
    * 7 7 8 8 *
    * * * * * *

    Note that the short row channels come after all the long row connections in
    the connections sequence.

    public attributes (in addition to init arguments and inherited attributes):
        unused_channels_xy
        long_out_channels_xy -- Output channels in the long rows.
    """

    # TODO: settle on 'long' or 'even' term?

    def __init__(self, in_channels_xy, field_channels_xy,
                 in_channel_dim=1, ignore_cover=False):
        """Calculate the connections.

        Keyword arguments:
        in_channels_xy -- 2-Tuple with number of input channels in the x- and
            y-direction (or a single number for both). This has to be
            specified, since the actual input is only one 1d array.
        field_channels_xy -- 2-Tuple with number of channels in each field in
            the x- and y-direction (or a single number for both).
            Must be even numbers.
        in_channel_dim -- Number of connections per input channel
        ignore_cover -- Boolean value defines if an
            Rectangular2dSwitchboardException is raised when the fields do not
            cover all input channels. Set this to True if you are willing to
            risk loosing input channels at the border.
        """
        in_channels_xy = to_2tuple(in_channels_xy)
        field_channels_xy = to_2tuple(field_channels_xy)
        ## count channels and stuff
        self.in_channels_xy = in_channels_xy
        self.field_channels_xy = field_channels_xy
        out_channel_dim = (in_channel_dim *
                           field_channels_xy[0] * field_channels_xy[1])
        ## check parameters for inconsistencies
        for i, name in enumerate(["x", "y"]):
            if field_channels_xy[i] % 2:
                err = ("%s_field_channels must be an even number, was %d" %
                       (name, field_channels_xy[i]))
                raise Rectangular2dSwitchboardException(err)
        field_spacing_xy = (field_channels_xy[0] // 2,
                            field_channels_xy[1] // 2)
        for i, name in enumerate(["x", "y"]):
            if (field_channels_xy[i] > in_channels_xy[i]):
                err = ("Number of field channels"
                       "exceeds the number of input channels in %s-direction. "
                       "This would lead to an empty connection list." %
                       name)
                raise Rectangular2dSwitchboardException(err)
        # number of output channels in x-direction
        xl = in_channels_xy[0] // field_channels_xy[0]
        x_unused_channels = in_channels_xy[0] - field_channels_xy[0]
        if x_unused_channels > 0:
            x_unused_channels %= field_spacing_xy[0]
        elif x_unused_channels < 0:
            x_unused_channels = in_channels_xy[0]
        if x_unused_channels and not ignore_cover:
            err = ("Channel fields do not "
                   "cover all input channels in x-direction.")
            raise Rectangular2dSwitchboardException(err)
        if (in_channels_xy[0] - xl * field_channels_xy[0]) >= (field_channels_xy[0] // 2):
            err = ("x short rows have same length as long rows.")
            raise Rectangular2dSwitchboardException(err)
        # number of output channels in y-direction
        yl = in_channels_xy[1] // field_channels_xy[1]
        y_unused_channels = in_channels_xy[1] - field_channels_xy[1]
        if y_unused_channels > 0:
            y_unused_channels %= field_spacing_xy[1]
        elif y_unused_channels < 0:
            y_unused_channels = in_channels_xy[1]
        if y_unused_channels and not ignore_cover:
            err = ("Channel fields do not "
                   "cover all input channels in y-direction.")
            raise Rectangular2dSwitchboardException(err)
        if ((in_channels_xy[1] - yl * field_channels_xy[1]) >=
            (field_channels_xy[1] // 2)):
            err = ("y short rows have same length as long rows.")
            raise Rectangular2dSwitchboardException(err)
        # TODO: add check against n+1/2 size, long line length equals short one
        ## end of parameters checks
        self.long_out_channels_xy = (xl, yl)
        self.unused_channels_xy = (x_unused_channels, y_unused_channels)
        out_channels = xl * yl + (xl-1) * (yl-1)
        in_trans = CoordinateTranslator(*in_channels_xy)
        connections = numx.zeros([out_channels * out_channel_dim],
                                 dtype=numx.int32)
        first_out_con = 0
        ## first create the even connections
        even_x_out_channels = in_channels_xy[0] // (2 * field_spacing_xy[0])
        even_y_out_channels = in_channels_xy[1] // (2 * field_spacing_xy[1])
        for y_out_chan in range(even_y_out_channels):
            for x_out_chan in range(even_x_out_channels):
                # inner loop over field
                x_start_chan = x_out_chan * (2 * field_spacing_xy[0])
                y_start_chan = y_out_chan * (2 * field_spacing_xy[1])
                for y_in_chan in range(y_start_chan,
                                    y_start_chan + self.field_channels_xy[1]):
                    for x_in_chan in range(x_start_chan,
                                    x_start_chan + self.field_channels_xy[0]):
                        first_in_con = (in_trans.image_to_index(
                                            x_in_chan, y_in_chan) *
                                        in_channel_dim)
                        connections[first_out_con:
                                    first_out_con + in_channel_dim] = \
                            list(range(first_in_con,
                                  first_in_con + in_channel_dim))
                        first_out_con += in_channel_dim
        ## create the uneven connections
        for y_out_chan in range(even_y_out_channels - 1):
            for x_out_chan in range(even_x_out_channels - 1):
                # inner loop over field
                x_start_chan = (x_out_chan * (2 * field_spacing_xy[0]) +
                                field_spacing_xy[0])
                y_start_chan = (y_out_chan * (2 * field_spacing_xy[1]) +
                                field_spacing_xy[1])
                for y_in_chan in range(y_start_chan,
                                       y_start_chan + self.field_channels_xy[1]):
                    for x_in_chan in range(x_start_chan,
                                       x_start_chan + self.field_channels_xy[0]):
                        first_in_con = (in_trans.image_to_index(
                                                    x_in_chan, y_in_chan) *
                                        in_channel_dim)
                        connections[first_out_con:
                                    first_out_con + in_channel_dim] = \
                            list(range(first_in_con,
                                  first_in_con + in_channel_dim))
                        first_out_con += in_channel_dim
        super(DoubleRect2dSwitchboard, self).__init__(
                                input_dim=in_channel_dim *
                                    in_channels_xy[0] * in_channels_xy[1],
                                connections=connections,
                                out_channel_dim=out_channel_dim,
                                in_channel_dim=in_channel_dim)


class DoubleRhomb2dSwitchboardException(SwitchboardException):
    """Exception for routing problems in the DoubleRhomb2dSwitchboard class."""
    pass


class DoubleRhomb2dSwitchboard(ChannelSwitchboard):
    """Rectangular lattice switchboard covering a rhombic lattice.

    All inner points of the rhombic lattice are covered twice. The rectangular
    fields are rotated by 45 degree.

    We assume that both the first and last row is a long row, e.g.

    *   *   *   *
      *   *   *
    *   *   *   *
      *   *   *
    *   *   *   *

    The incoming data is expected to contain the long rows first, then
    the short rows.

    The alignment of the first field is chosen to minimize cutaway.

    public attributes (in addition to init arguments and inherited attributes):
        out_channels_xy
    """

    def __init__(self, long_in_channels_xy,
                 diag_field_channels, in_channel_dim=1):
        """Calculate the connections.

        Note that the incoming data will be interpreted as a rhombic grid,
        as it is produced by DoubleRect2dSwitchboard.

        Keyword arguments:
        long_in_channels_xy -- 2-Tuple with number of long input channels in
            the x- and y-direction (or a single number for both).
        diag_field_channels -- Field edge size (before the rotation).
        in_channel_dim -- Number of connections per input channel
        """
        long_in_channels_xy = to_2tuple(long_in_channels_xy)
        self.long_in_channels_xy = long_in_channels_xy
        if long_in_channels_xy[0] < long_in_channels_xy[1]:
            started_in_short = 1
        else:
            started_in_short = 0
        ## check parameters for inconsistencies ##
        if diag_field_channels % 2:
            err = ("diag_field_channels must be even (for double cover)")
            raise DoubleRhomb2dSwitchboardException(err)
        self.diag_field_channels = diag_field_channels
        # helper variables for the field range
        _x_chan_field_range = (long_in_channels_xy[0] -
                               (1 - started_in_short) - diag_field_channels)
        _y_chan_field_range = (long_in_channels_xy[1] - started_in_short -
                               diag_field_channels)
        if (_x_chan_field_range % (diag_field_channels // 2) or
            _x_chan_field_range < 0):
            err = ("diag_field_channels value is not compatible with "
                   "long_in_channels_xy[0]")
            raise DoubleRhomb2dSwitchboardException(err)
        if (_y_chan_field_range % (diag_field_channels // 2) or
            _y_chan_field_range < 0):
            err = ("diag_field_channels value is not compatible with "
                   "long_in_channels_xy[1]")
            raise DoubleRhomb2dSwitchboardException(err)
        ## count channels and stuff
        self.in_channel_dim = in_channel_dim
        input_dim = ((2 * long_in_channels_xy[0] * long_in_channels_xy[1]
                     - long_in_channels_xy[0] - long_in_channels_xy[1] + 1) *
                     in_channel_dim)
        out_channel_dim = in_channel_dim * diag_field_channels**2
        x_out_channels = (2 * _x_chan_field_range // diag_field_channels + 1)
        y_out_channels = (2 * _y_chan_field_range // diag_field_channels + 1)
        self.out_channels_xy = (x_out_channels, y_out_channels)
        ## prepare iteration over fields
        long_in_trans = CoordinateTranslator(*long_in_channels_xy)
        short_in_trans = CoordinateTranslator(long_in_channels_xy[0] - 1,
                                              long_in_channels_xy[1] - 1)
        short_in_offset = long_in_channels_xy[0] * long_in_channels_xy[1]
        connections = numx.zeros([x_out_channels * y_out_channels *
                                  out_channel_dim], dtype=numx.int32)
        first_out_con = 0
        for y_out_chan in range(y_out_channels):
            for x_out_chan in range(x_out_channels):
                # inner loop over perceptive field
                x_start_chan = (1 + x_out_chan) * diag_field_channels // 2
                y_start_chan = y_out_chan * diag_field_channels
                # set the initial field offset to minimize edge loss
                x_start_chan -= started_in_short
                y_start_chan += started_in_short
                # iterate over both long and short rows
                for iy, y_in_chan in enumerate(range(y_start_chan,
                                y_start_chan + (2 * diag_field_channels - 1))):
                    # half width of the field in the given row
                    if iy <= (diag_field_channels - 1):
                        field_width = iy + 1
                    else:
                        field_width = (diag_field_channels - 1 -
                                       (iy % diag_field_channels))
                    for x_in_chan in range(x_start_chan - field_width // 2,
                                           x_start_chan + field_width // 2
                                                        + field_width % 2):
                        # array index of the first input connection
                        # for this input channel
                        if not y_in_chan % 2:
                            if started_in_short:
                                x_in_chan += 1
                            first_in_con = (
                                long_in_trans.image_to_index(
                                                x_in_chan, y_in_chan // 2) *
                                                        self.in_channel_dim)
                        else:
                            first_in_con = (
                                (short_in_trans.image_to_index(
                                                x_in_chan, y_in_chan // 2)
                                 + short_in_offset) * self.in_channel_dim)
                        connections[first_out_con:
                                    first_out_con + self.in_channel_dim] = \
                            list(range(first_in_con,
                                  first_in_con + self.in_channel_dim))
                        first_out_con += self.in_channel_dim
        super(DoubleRhomb2dSwitchboard, self).__init__(
                                        input_dim=input_dim,
                                        connections=connections,
                                        out_channel_dim=out_channel_dim,
                                        in_channel_dim=in_channel_dim)


# utility class for Rectangular2dSwitchboard

class CoordinateTranslator(object):
    """Translate between image (PIL) and numpy array coordinates.

    PIL image coordinates go from 0..width-1 . The first coordinate is x.
    Array coordinates also start from 0, but the first coordinate is the row.
    As depicted below we have x = column, y = row. The entry index numbers are
    also shown.

      +------> x
      | 1 2
      | 3 4
    y v

    array[y][x]
    """

    def __init__(self, x_image_dim, y_image_dim):
        self.x_image_dim = x_image_dim
        self.y_image_dim = y_image_dim
        self._max_index = x_image_dim * y_image_dim - 1

    def image_to_array(self, x, y):
        return y, x

    def image_to_index(self, x, y):
        if not 0 <= x < self.x_image_dim:
            raise Exception("x coordinate %d is outside the valid range." % x)
        if not 0 <= y < self.y_image_dim:
            raise Exception("y coordinate %d is outside the valid range." % y)
        return y * self.x_image_dim + x

    def array_to_image(self, row, col):
        return col, row

    def array_to_index(self, row, col):
        if not 0 <= row < self.y_image_dim:
            raise Exception("row index %d is outside the valid range." % row)
        if not 0 <= col < self.x_image_dim:
            raise Exception("column index %d is outside the valid range." %
                            col)
        return row * self.x_image_dim + col

    def index_to_array(self, index):
        if not 0 <= index <= self._max_index:
            raise Exception("index %d is outside the valid range." %
                            index)
        return index // self.x_image_dim, index % self.x_image_dim

    def index_to_image(self, index):
        if not 0 <= index <= self._max_index:
            raise Exception("index %d is outside the valid range." %
                            index)
        return index % self.x_image_dim, index // self.x_image_dim



"""
This is the MDP package for parallel processing.

It is designed to work with nodes for which a large part of the
computation is embaressingly parallel (like in
:class:`~mdp.nodes.PCANode`). The hinet package is also fully
supported, i.e., there are parallel versions of all hinet nodes.

This package consists of two decoupled parts. The first part consists
of parallel versions of the familiar MDP structures (nodes and
flows). At the top there is the :class:`~ParallelFlow`, which
generates tasks that are processed in parallel (this can be done
automatically in the train or execute methods).

The second part consists of the schedulers. They take tasks and
process them in a more or less parallel way (e.g. in multiple
processes). So they are designed to deal with the more technical
aspects of the parallelization, but do not have to know anything about
flows or nodes.
"""


from .scheduling import (
    ResultContainer, ListResultContainer, OrderedResultContainer, TaskCallable,
    SqrTestCallable, SleepSqrTestCallable, TaskCallableWrapper, Scheduler,
    cpu_count, MDPVersionCallable
)
from .process_schedule import ProcessScheduler
from .thread_schedule import ThreadScheduler
from .parallelnodes import (
    ParallelExtensionNode, NotForkableParallelException, JoinParallelException,
    ParallelPCANode, ParallelSFANode, ParallelFDANode, ParallelHistogramNode
)
from .parallelclassifiers import (
    ParallelGaussianClassifier, ParallelNearestMeanClassifier,
    ParallelKNNClassifier
)
from .parallelflows import (
    _purge_flownode, FlowTaskCallable, FlowTrainCallable, FlowExecuteCallable,
    TrainResultContainer, ExecuteResultContainer,
    ParallelFlowException, NoTaskException,
    ParallelFlow, ParallelCheckpointFlow
)
from .parallelhinet import (
    ParallelFlowNode, ParallelLayer, ParallelCloneLayer
)

from mdp import config
from mdp.utils import fixup_namespace

if config.has_parallel_python:
    from . import pp_support

# Note: the modules with the actual extension node classes are still available

__all__ = [
    "ResultContainer", "ListResultContainer",
    "OrderedResultContainer", "TaskCallable", "SqrTestCallable",
    "SleepSqrTestCallable", "TaskCallableWrapper", "Scheduler",
    "ProcessScheduler", "ThreadScheduler",
    "ParallelExtensionNode", "JoinParallelException",
    "NotForkableParallelException",
    "ParallelSFANode", "ParallelSFANode", "ParallelFDANode",
    "ParallelHistogramNode",
    "FlowTaskCallable", "FlowTrainCallable", "FlowExecuteCallable",
    "ExecuteResultContainer", "TrainResultContainer", "ParallelFlowException",
    "NoTaskException",
    "ParallelFlow", "ParallelCheckpointFlow",
    "ParallelFlowNode", "ParallelLayer", "ParallelCloneLayer"]

import sys as _sys
fixup_namespace(__name__, __all__,
                ('scheduling',
                 'process_schedule',
                 'thread_schedule',
                 'parallelnodes',
                 'parallelflows',
                 'parallelhinet',
                 'parallelclassifiers',
                 'config',
                 'fixup_namespace'
                 ))



from future import standard_library
standard_library.install_aliases()
from builtins import range
from builtins import object

import tempfile
import pickle
import pickle
import os
from ._tools import *

uniform = numx_rand.random

def _get_default_flow(flow_class=mdp.Flow, node_class=BogusNode):
    flow = flow_class([node_class(),node_class(),node_class()])
    return flow

# CheckpointFunction used in testCheckpointFunction
class _CheckpointCollectFunction(mdp.CheckpointFunction):
    def __init__(self):
        self.classes = []

    # collect the classes of the nodes it checks
    def __call__(self, node):
        self.classes.append(node.__class__)


def testFlow():
    inp = numx.ones((100,3))
    flow = _get_default_flow()
    for i in range(len(flow)):
        assert not flow.flow[i].is_training(), \
               'Training of node #%d has not been closed.' % i

    out = flow(inp)
    assert_array_equal(out,(2**len(flow))*inp)
    rec = flow.inverse(out)
    assert_array_equal(rec,inp)

def testFlow_copy():
    dummy_list = [1,2,3]
    flow = _get_default_flow()
    flow[0].dummy_attr = dummy_list
    copy_flow = flow.copy()
    assert flow[0].dummy_attr == copy_flow[0].dummy_attr, \
           'Flow copy method did not work'
    copy_flow[0].dummy_attr[0] = 10
    assert flow[0].dummy_attr != copy_flow[0].dummy_attr, \
           'Flow copy method did not work'

def test_Flow_copy_with_lambda():
    generic_node = mdp.Node()
    generic_node.lambda_function = lambda: 1
    generic_flow = mdp.Flow([generic_node])
    generic_flow.copy()
    
def testFlow_save():
    dummy_list = [1,2,3]
    flow = _get_default_flow()
    flow[0].dummy_attr = dummy_list
    # test string save
    copy_flow_pic = flow.save(None)
    copy_flow = pickle.loads(copy_flow_pic)
    assert flow[0].dummy_attr == copy_flow[0].dummy_attr, \
           'Flow save (string) method did not work'
    copy_flow[0].dummy_attr[0] = 10
    assert flow[0].dummy_attr != copy_flow[0].dummy_attr, \
           'Flow save (string) method did not work'
    # test file save
    dummy_file = tempfile.mktemp(prefix='MDP_', suffix=".pic",
                                 dir=py.test.mdp_tempdirname)
    flow.save(dummy_file, protocol=1)
    dummy_file = open(dummy_file, 'rb')
    copy_flow = pickle.load(dummy_file)
    assert flow[0].dummy_attr == copy_flow[0].dummy_attr, \
           'Flow save (file) method did not work'
    copy_flow[0].dummy_attr[0] = 10
    assert flow[0].dummy_attr != copy_flow[0].dummy_attr, \
           'Flow save (file) method did not work'

def testFlow_container_privmethods():
    mat,mix,inp = get_random_mix(mat_dim=(100,3))
    flow = _get_default_flow()
    # test __len__
    assert_equal(len(flow), len(flow.flow))
    # test __?etitem__, integer key
    for i in range(len(flow)):
        assert flow[i]==flow.flow[i], \
               '__getitem__  returned wrong node %d' % i
        new_node = BogusNode()
        flow[i] = new_node
        assert flow[i]==new_node, '__setitem__ did not set node %d' % i
    # test __?etitem__, normal slice -> this fails for python < 2.2 and
    # if Flow is a subclassed from builtin 'list'
    flowslice = flow[0:2]
    assert isinstance(flowslice,mdp.Flow), \
           '__getitem__ slice is not a Flow instance'
    assert len(flowslice) == 2, '__getitem__ returned wrong slice size'
    new_nodes_list = [BogusNode(), BogusNode()]
    flow[:2] = new_nodes_list
    assert (flow[0] == new_nodes_list[0]) and \
           (flow[1] == new_nodes_list[1]), '__setitem__ did not set slice'
    # test__?etitem__, extended slice
    flowslice = flow[:2:1]
    assert isinstance(flowslice,mdp.Flow), \
           '__getitem__ slice is not a Flow instance'
    assert len(flowslice) == 2, '__getitem__ returned wrong slice size'
    new_nodes_list = [BogusNode(), BogusNode()]
    flow[:2:1] = new_nodes_list
    assert (flow[0] == new_nodes_list[0]) and \
           (flow[1] == new_nodes_list[1]), '__setitem__ did not set slice'
    # test __delitem__, integer key
    copy_flow = mdp.Flow(flow[:])
    del copy_flow[0]
    assert len(copy_flow) == len(flow)-1, '__delitem__ did not del'
    for i in range(len(copy_flow)):
        assert copy_flow[i] == flow[i+1], '__delitem__ deleted wrong node'
    # test __delitem__, normal slice
    copy_flow = mdp.Flow(flow[:])
    del copy_flow[:2]
    assert len(copy_flow) == len(flow)-2, \
           '__delitem__ did not del normal slice'
    assert copy_flow[0] == flow[2], \
           '__delitem__ deleted wrong normal slice'
    # test __delitem__, extended slice
    copy_flow = mdp.Flow(flow[:])
    del copy_flow[:2:1]
    assert len(copy_flow) == len(flow)-2, \
           '__delitem__ did not del extended slice'
    assert copy_flow[0] == flow[2], \
           '__delitem__ deleted wrong extended slice'
    # test __add__
    newflow = flow + flow
    assert len(newflow) == len(flow)*2, '__add__ did not work'

def testFlow_container_listmethods():
    # for all methods try using a node with right dimensionality
    # and one with wrong dimensionality
    flow = _get_default_flow()
    length = len(flow)
    # we test __contains__ and __iter__ with the for loop
    for node in flow:
        node.input_dim = 10
        node.output_dim = 10
    # append
    newnode = BogusNode(input_dim=10, output_dim=10)
    flow.append(newnode)
    assert_equal(len(flow), length+1)
    length = len(flow)
    try:
        newnode = BogusNode(input_dim=11)
        flow.append(newnode)
        raise Exception('flow.append appended inconsistent node')
    except ValueError:
        assert_equal(len(flow), length)
    # extend
    newflow = flow.copy()
    flow.extend(newflow)
    assert_equal(len(flow), 2*length)
    length = len(flow)
    try:
        newflow = _get_default_flow()
        for idx in range(len(newflow)):
            if idx == 0:
                newflow[idx].input_dim = 11
            else:
                newflow[idx].input_dim = 10
            newflow[idx].output_dim = 10
        flow.extend(newflow)
        raise Exception('flow.extend appended inconsistent flow')
    except ValueError:
        assert_equal(len(flow), length)
    # insert
    newnode = BogusNode(input_dim=10, output_dim=None)
    flow.insert(2, newnode)
    assert_equal(len(flow), length+1)
    length = len(flow)
    try:
        newnode = BogusNode(output_dim=11)
        flow.insert(2, newnode)
        raise Exception('flow.insert inserted inconsistent node')
    except ValueError:
        assert_equal(len(flow), length)
    # pop
    oldnode = flow[5]
    popnode = flow.pop(5)
    assert oldnode == popnode, 'flow.pop popped wrong node out'
    assert_equal(len(flow), length-1)
    # pop - test Flow._check_nodes_consistency
    flow = _get_default_flow() + _get_default_flow()
    length = len(flow)
    flow[3].output_dim = 2
    flow[4].input_dim = 2
    flow[4].output_dim = 3
    flow[5].input_dim = 3
    flow._check_nodes_consistency(flow.flow)
    try:
        nottobepopped = flow.pop(4)
        raise Exception('flow.pop left inconsistent flow')
    except ValueError:
        assert_equal(len(flow), length)

def testFlow_append_node_copy():
    # when appending a node to a flow,
    # we don't want the flow to be a copy!
    node1 = BogusNode()
    node2 = BogusNode()
    flow = mdp.Flow([node1])
    flow += node2
    assert flow[0] is node1
    
def testFlow_iadd():
    # check that in-place adding to flow does not return new flow
    node1 = BogusNode()
    node2 = BogusNode()
    node3 = BogusNode()
    flow = mdp.Flow([node1])
    oldflow = flow
    flow += node2
    assert oldflow is flow
    flow += mdp.Flow([node3])
    assert oldflow is flow

def testFlow_as_sum_of_nodes():
    node1 = BogusNode()
    node2 = BogusNode()
    flow = node1+node2
    assert type(flow) is mdp.Flow
    assert len(flow) == 2
    node3 = BogusNode()
    flow = node1+node2+node3
    assert type(flow) is mdp.Flow
    assert len(flow) == 3
    node4 = BogusNode()
    flow = node4 + flow
    assert type(flow) is mdp.Flow
    assert len(flow) == 4

def testFlowWrongItarableException():
    samples = mdp.numx_rand.random((100,10))
    labels = mdp.numx.arange(100)
    flow = mdp.Flow([mdp.nodes.PCANode(), mdp.nodes.FDANode()])
    try:
        flow.train([[samples], [samples, labels]])
        # correct line would be (note the second iterable):
        #    flow.train([[[samples]], [[samples, labels]]])
        # should trigger exception for missing train argument for FDANode
        err = "Flow did not raise FlowException for wrong iterable."
        raise Exception(err)
    except mdp.FlowException:
        pass
    try:
        # try to give one argument too much!
        flow.train([[[samples]], [[samples, labels, labels]]])
        err = "Flow did not raise FlowException for wrong iterable."
        raise Exception(err)
    except mdp.FlowException:
        pass


def testCheckpointFlow():
    lst = []
    # checkpoint function, it collects a '1' for each call
    def cfunc(node, lst = lst):
        lst.append(1)
    mat,mix,inp = get_random_mix(mat_dim=(100,3))
    flow = _get_default_flow(flow_class = mdp.CheckpointFlow,
                                  node_class = BogusNodeTrainable)
    flow.train(inp, cfunc)
    #
    assert len(lst)==len(flow), \
           'The checkpoint function has been called %d times instead of %d times.' % (len(lst), len(flow))

def testCheckpointFunction():
    cfunc = _CheckpointCollectFunction()
    mat,mix,inp = get_random_mix(mat_dim=(100,3))
    flow = _get_default_flow(flow_class = mdp.CheckpointFlow,
                                  node_class = BogusNodeTrainable)
    flow.train(inp, cfunc)
    #
    for i in range(len(flow)):
        assert flow[i].__class__==cfunc.classes[i], 'Wrong class collected'

def testCrashRecovery():
    flow = mdp.Flow([BogusExceptNode()])
    flow.set_crash_recovery(1)
    try:
        flow.train(mdp.numx.zeros((1,2), 'd'))
    except Exception as e:
        assert isinstance(e,mdp.FlowExceptionCR)
        with open(e.filename, 'rb') as fl:
            pic_flow = pickle.load(fl)
        os.remove(e.filename)
        assert flow[0].bogus_attr == pic_flow[0].bogus_attr
    flow.set_crash_recovery(0)
    try:
        flow.execute([None])
    except Exception as e:
        assert isinstance(e,mdp.FlowExceptionCR)
        assert not hasattr(e,'filename')

def testCrashRecoveryException():
    a = 3
    try:
        raise mdp.CrashRecoveryException('bogus errstr', a, Exception())
    except mdp.CrashRecoveryException as e:
        filename1 = e.dump()
        filename2 = e.dump(tempfile.mkstemp(prefix='MDP_',
                                            dir=py.test.mdp_tempdirname)[1])
        assert isinstance(e.parent_exception, Exception)

    for fname in filename1, filename2:
        fl = open(fname, 'rb')
        obj = pickle.load(fl)
        fl.close()
        try:
            os.remove(fname)
        except Exception:
            pass
        assert obj == a

def testMultiplePhases():
    # test basic multiple phase sequence
    flow = mdp.Flow([BogusMultiNode()])
    flow.train(mdp.numx.zeros((1,2), 'd'))
    assert flow[0].visited == [1,2,3,4]
    # try to use an iterable to train it, check for rewinds
    class TestIterable(object):
        def __init__(self):
            self.used = 0
        def __iter__(self):
            self.used += 1
            yield mdp.numx.zeros((1,2), 'd')
    flow = mdp.Flow([BogusMultiNode()])
    iterable = TestIterable()
    flow.train([iterable])
    assert iterable.used == 2
    # should not work with an iterator
    def testgenerator():
        yield mdp.numx.zeros((1,2), 'd')
    flow = mdp.Flow([BogusMultiNode()])
    try:
        flow.train([testgenerator()])
        raise Exception('Expected mdp.FlowException')
    except mdp.FlowException:
        pass



from builtins import range
from ._tools import *

import mdp.parallel as parallel
n = numx

def test_process_scheduler_shutdown():
    """Test that we can properly shutdown the subprocesses"""
    scheduler = parallel.ProcessScheduler(verbose=False,
                                          n_processes=1,
                                          source_paths=None,
                                          cache_callable=False)
    scheduler.shutdown()

def test_process_scheduler_order():
    """Test the correct result order in process scheduler."""
    scheduler = parallel.ProcessScheduler(verbose=False,
                                          n_processes=3,
                                          source_paths=None)
    max_i = 8
    for i in range(max_i):
        scheduler.add_task((n.arange(0,i+1), (max_i-1-i)*1.0/4),
                           parallel.SleepSqrTestCallable())
    results = scheduler.get_results()
    scheduler.shutdown()
    # check result
    results = n.concatenate(results)
    assert n.all(results ==
                     n.concatenate([n.arange(0,i+1)**2
                                    for i in range(max_i)]))

def test_process_scheduler_no_cache():
    """Test process scheduler with caching turned off."""
    scheduler = parallel.ProcessScheduler(verbose=False,
                                          n_processes=2,
                                          source_paths=None,
                                          cache_callable=False)
    for i in range(8):
        scheduler.add_task(i, parallel.SqrTestCallable())
    results = scheduler.get_results()
    scheduler.shutdown()
    # check result
    results = n.array(results)
    assert n.all(results == n.array([0,1,4,9,16,25,36,49]))
    
def test_process_scheduler_manager():
    """Test process scheduler with context manager itnerface."""
    with parallel.ProcessScheduler(n_processes=2,
                                   source_paths=None) as scheduler:
        for i in range(8):
            scheduler.add_task(i, parallel.SqrTestCallable())
        results = scheduler.get_results()
    # check result
    results = n.array(results)
    assert n.all(results == n.array([0,1,4,9,16,25,36,49]))

def test_process_scheduler_flow():
    """Test process scheduler with real Nodes."""
    precision = 6
    node1 = mdp.nodes.PCANode(output_dim=20)
    node2 = mdp.nodes.PolynomialExpansionNode(degree=1)
    node3 = mdp.nodes.SFANode(output_dim=10)
    flow = mdp.parallel.ParallelFlow([node1, node2, node3])
    parallel_flow = mdp.parallel.ParallelFlow(flow.copy()[:])
    input_dim = 30
    scales = n.linspace(1, 100, num=input_dim)
    scale_matrix = mdp.numx.diag(scales)
    train_iterables = [n.dot(mdp.numx_rand.random((5, 100, input_dim)),
                             scale_matrix)
                       for _ in range(3)]
    x = mdp.numx.random.random((10, input_dim))
    with parallel.ProcessScheduler(verbose=False,
                                   n_processes=3,
                                   source_paths=None) as scheduler:
        parallel_flow.train(train_iterables, scheduler=scheduler)
        # test that parallel execution works as well
        # note that we need more chungs then processes to test caching
        parallel_flow.execute([x for _ in range(8)], scheduler=scheduler)
    # compare to normal flow
    flow.train(train_iterables)
    assert parallel_flow[0].tlen == flow[0].tlen
    y1 = flow.execute(x)
    y2 = parallel_flow.execute(x)
    assert_array_almost_equal(abs(y1), abs(y2), precision)

def test_process_scheduler_mdp_version():
    """Test that we are running the same mdp in subprocesses"""
    scheduler = parallel.ProcessScheduler(verbose=False,
                                          n_processes=2,
                                          source_paths=None,
                                          cache_callable=False)
    for i in range(2):
        scheduler.add_task(i, parallel.MDPVersionCallable())
    out = scheduler.get_results()
    scheduler.shutdown()
    # check that we get 2 identical dictionaries
    assert out[0] == out[1], 'Subprocesses did not run '\
        'the same MDP as the parent:\n%s\n--\n%s'%(out[0], out[1])



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from mdp import numx
from mdp.utils import mult

random = mdp.numx_rand.random
randn = mdp.numx_rand.randn
exp = mdp.numx.exp

# TODO: does it make sense to define the inverse of RBMNode as sampling
#    from the visible layer given an hidden state?

# this and the other replication functions should go in mdp.utils
def rrep(x, n):
    """Replicate x n-times on a new last dimension"""
    shp = x.shape + (1,)
    return x.reshape(shp).repeat(n, axis=-1)


class RBMNode(mdp.Node):
    """Restricted Boltzmann Machine node. An RBM is an undirected
    probabilistic network with binary variables. The graph is
    bipartite into observed (*visible*) and hidden (*latent*) variables.

    By default, the ``execute`` method returns the *probability* of
    one of the hiden variables being equal to 1 given the input.

    Use the ``sample_v`` method to sample from the observed variables
    given a setting of the hidden variables, and ``sample_h`` to do the
    opposite. The ``energy`` method can be used to compute the energy
    of a given setting of all variables.

    The network is trained by Contrastive Divergence, as described in
    Hinton, G. E. (2002). Training products of experts by minimizing
    contrastive divergence. Neural Computation, 14(8):1711-1800

    **Internal variables of interest**

      ``self.w``
          Generative weights between hidden and observed variables

      ``self.bv``
          bias vector of the observed variables

      ``self.bh``
          bias vector of the hidden variables

    For more information on RBMs, see
    Geoffrey E. Hinton (2007) Boltzmann machine. Scholarpedia, 2(5):1668
    """

    def __init__(self, hidden_dim, visible_dim = None, dtype = None):
        """
        :Parameters:
          hidden_dim
            number of hidden variables
          visible_dim
            number of observed variables
        """
        super(RBMNode, self).__init__(visible_dim, hidden_dim, dtype)
        self._initialized = False

    def _init_weights(self):
        # weights and biases are initialized to small random values to
        # break the simmetry that might lead to degenerate solutions during
        # learning
        self._initialized = True

        # weights
        self.w = self._refcast(randn(self.input_dim, self.output_dim)*0.1)
        # bias on the visibile (input) units
        self.bv = self._refcast(randn(self.input_dim)*0.1)
        # bias on the hidden (output) units
        self.bh = self._refcast(randn(self.output_dim)*0.1)

        # delta w, bv, bh used for momentum term
        self._delta = (0., 0., 0.)

    def _sample_h(self, v):
        # returns P(h=1|v,W,b) and a sample from it
        probs = old_div(1.,(1. + exp(-self.bh - mult(v, self.w))))
        h = (probs > random(probs.shape)).astype(self.dtype)
        return probs, h

    def _sample_v(self, h):
        # returns  P(v=1|h,W,b) and a sample from it
        probs = old_div(1.,(1. + exp(-self.bv - mult(h, self.w.T))))
        v = (probs > random(probs.shape)).astype(self.dtype)
        return probs, v

    def _train(self, v, n_updates=1, epsilon=0.1, decay=0., momentum=0.,
               update_with_ph=True, verbose=False):
        """Update the internal structures according to the input data `v`.
        The training is performed using Contrastive Divergence (CD).

        :Parameters:
          v
            a binary matrix having different variables on different columns
            and observations on the rows
          n_updates
            number of CD iterations. Default value: 1
          epsilon
            learning rate. Default value: 0.1
          decay
            weight decay term. Default value: 0.
          momentum
            momentum term. Default value: 0.
          update_with_ph
            In his code, G.Hinton updates the hidden biases using the
            probability of the hidden unit activations instead of a
            sample from it. This is in order to speed up sequential
            learning of RBMs. Set this to False to use the samples instead.
        """
        if not self._initialized:
            self._init_weights()

        # useful quantities
        n = v.shape[0]
        w, bv, bh = self.w, self.bv, self.bh

        # old gradients for momentum term
        dw, dbv, dbh = self._delta

        # first update of the hidden units for the data term
        ph_data, h_data = self._sample_h(v)
        # n updates of both v and h for the model term
        h_model = h_data.copy()
        for i in range(n_updates):
            pv_model, v_model = self._sample_v(h_model)
            ph_model, h_model = self._sample_h(v_model)

        # update w
        data_term = mult(v.T, ph_data)
        model_term = mult(v_model.T, ph_model)
        dw = momentum*dw + epsilon*(old_div((data_term - model_term),n) - decay*w)
        w += dw

        # update bv
        data_term = v.sum(axis=0)
        model_term = v_model.sum(axis=0)
        dbv = momentum*dbv + epsilon*(old_div((data_term - model_term),n))
        bv += dbv

        # update bh
        if update_with_ph:
            data_term = ph_data.sum(axis=0)
            model_term = ph_model.sum(axis=0)
        else:
            data_term = h_data.sum(axis=0)
            model_term = h_model.sum(axis=0)
        dbh = momentum*dbh + epsilon*(old_div((data_term - model_term),n))
        bh += dbh

        self._delta = (dw, dbv, dbh)
        self._train_err = float(((v-v_model)**2.).sum())

        if verbose:
            print('training error', old_div(self._train_err,v.shape[0]))
            ph, h = self._sample_h(v)
            print('energy', self._energy(v, ph).sum())

    def _stop_training(self):
        #del self._delta
        #del self._train_err
        pass

    # execution methods

    @staticmethod
    def is_invertible():
        return False

    def _pre_inversion_checks(self, y):
        self._if_training_stop_training()
        # control the dimension of y
        self._check_output(y)

    def sample_h(self, v):
        """Sample the hidden variables given observations v.

        :Returns: a tuple ``(prob_h, h)``, where ``prob_h[n,i]`` is the
          probability that variable ``i`` is one given the observations
          ``v[n,:]``, and ``h[n,i]`` is a sample from the posterior probability.
        """
        self._pre_execution_checks(v)
        return self._sample_h(v)

    def sample_v(self, h):
        """Sample the observed variables given hidden variable state h.

        :Returns: a tuple ``(prob_v, v)``, where ``prob_v[n,i]`` is the
          probability that variable ``i`` is one given the hidden
          variables ``h[n,:]``, and ``v[n,i]`` is a sample from that
          conditional probability.
        """
        self._pre_inversion_checks(h)
        return self._sample_v(h)

    def _energy(self, v, h):
        return (-mult(v, self.bv) - mult(h, self.bh) -
                (mult(v, self.w)*h).sum(axis=1))

    def energy(self, v, h):
        """Compute the energy of the RBM given observed variables state `v` and
        hidden variables state `h`.
        """
        return self._energy(v, h)

    def _execute(self, v, return_probs=True):
        """If `return_probs` is True, returns the probability of the
        hidden variables h[n,i] being 1 given the observations v[n,:].
        If `return_probs` is False, return a sample from that probability.
        """
        probs, h = self._sample_h(v)
        if return_probs:
            return probs
        else:
            return h

class RBMWithLabelsNode(RBMNode):
    """Restricted Boltzmann Machine with softmax labels. An RBM is an
    undirected probabilistic network with binary variables. In this
    case, the node is partitioned into a set of observed (*visible*)
    variables, a set of hidden (*latent*) variables, and a set of
    label variables (also observed), only one of which is active at
    any time. The node is able to learn associations between the
    visible variables and the labels.

    By default, the ``execute`` method returns the *probability* of
    one of the hiden variables being equal to 1 given the input.

    Use the ``sample_v`` method to sample from the observed variables
    (visible and labels) given a setting of the hidden variables, and
    ``sample_h`` to do the opposite. The ``energy`` method can be used
    to compute the energy of a given setting of all variables.

    The network is trained by Contrastive Divergence, as described in
    Hinton, G. E. (2002). Training products of experts by minimizing
    contrastive divergence. Neural Computation, 14(8):1711-1800

    Internal variables of interest:

      ``self.w``
          Generative weights between hidden and observed variables

      ``self.bv``
          bias vector of the observed variables

      ``self.bh``
          bias vector of the hidden variables

    For more information on RBMs with labels, see
    
      * Geoffrey E. Hinton (2007) Boltzmann machine. Scholarpedia, 2(5):1668.
      * Hinton, G. E, Osindero, S., and Teh, Y. W. (2006). A fast learning
        algorithm for deep belief nets. Neural Computation, 18:1527-1554.
    """

    def __init__(self, hidden_dim, labels_dim, visible_dim=None, dtype=None):
        super(RBMWithLabelsNode, self).__init__(None, None, dtype)

        self._labels_dim = labels_dim
        if visible_dim is not None:
            self.input_dim = visible_dim+labels_dim

        self.output_dim = hidden_dim
        self._initialized = False

    def _set_input_dim(self, n):
        self._input_dim = n
        self._visible_dim = n - self._labels_dim

    def _sample_v(self, h, sample_l=False, concatenate=True):
        # returns  P(v=1|h,W,b), a sample from it, P(l=1|h,W,b),
        # and a sample from it

        ldim, vdim = self._labels_dim, self._visible_dim

        # activation
        a = self.bv + mult(h, self.w.T)
        av, al = a[:, :vdim], a[:, vdim:]

        # ## visible units: logistic activation
        probs_v = old_div(1.,(1. + exp(-av)))
        v = (probs_v > random(probs_v.shape)).astype('d')

        # ## label units: softmax activation
        # subtract maximum to regularize exponent
        exponent = al - rrep(al.max(axis=1), ldim)
        probs_l = exp(exponent)
        probs_l /= rrep(probs_l.sum(axis=1), ldim)

        if sample_l:
            # ?? todo: I'm sure this can be optimized
            l = numx.zeros((h.shape[0], ldim))
            for t in range(h.shape[0]):
                l[t, :] = mdp.numx_rand.multinomial(1, probs_l[t, :])
        else:
            l = probs_l.copy()

        if concatenate:
            probs = numx.concatenate((probs_v, probs_l), axis=1)
            x = numx.concatenate((v, l), axis=1)
            return probs, x
        else:
            return probs_v, probs_l, v, l

    # execution methods

    def sample_h(self, v, l):
        """Sample the hidden variables given observations `v` and labels `l`.

        :Returns: a tuple ``(prob_h, h)``, where ``prob_h[n,i]`` is the
          probability that variable ``i`` is one given the observations
          ``v[n,:]`` and the labels ``l[n,:]``, and ``h[n,i]`` is a sample
          from the posterior probability."""

        x = numx.concatenate((v, l), axis=1)
        self._pre_execution_checks(x)
        return self._sample_h(x)

    def sample_v(self, h):
        """Sample the observed variables given hidden variable state `h`.

        :Returns: a tuple ``(prob_v, probs_l, v, l)``, where ``prob_v[n,i]``
          is the probability that the visible variable ``i`` is one given
          the hidden variables ``h[n,:]``, and ``v[n,i]`` is a sample from
          that conditional probability. ``prob_l`` and ``l`` have similar
          interpretations for the label variables. Note that the labels are
          activated using a softmax function, so that only one label can be
          active at any time.
        """
        self._pre_inversion_checks(h)

        probs_v, probs_l, v, l = self._sample_v(h, sample_l=True,
                                                concatenate=False)
        return probs_v, probs_l, v, l

    def energy(self, v, h, l):
        """Compute the energy of the RBM given observed variables state `v`
        and `l`, and hidden variables state `h`."""

        x = numx.concatenate((v, l), axis=1)
        return self._energy(x, h)

    def execute(self, v, l, return_probs = True):
        """If `return_probs` is True, returns the probability of the
        hidden variables h[n,i] being 1 given the observations v[n,:]
        and l[n,:].  If `return_probs` is False, return a sample from
        that probability.
        """
        x = numx.concatenate((v, l), axis=1)
        self._pre_execution_checks(x)

        probs, h = self._sample_h(self._refcast(x))
        if return_probs:
            return probs
        else:
            return h

    @staticmethod
    def is_invertible():
        return False

    def train(self, v, l, n_updates=1, epsilon=0.1, decay=0., momentum=0.,
              verbose=False):
        """Update the internal structures according to the visible data `v`
        and the labels `l`.
        The training is performed using Contrastive Divergence (CD).

        :Parameters:
          v
            a binary matrix having different variables on different columns
            and observations on the rows
          l
            a binary matrix having different variables on different columns
            and observations on the rows. Only one value per row should be 1.
          n_updates
            number of CD iterations. Default value: 1
          epsilon
            learning rate. Default value: 0.1
          decay
            weight decay term. Default value: 0.
          momentum
            momentum term. Default value: 0.
        """

        if not self.is_training():
            errstr = "The training phase has already finished."
            raise mdp.TrainingFinishedException(errstr)

        x = numx.concatenate((v, l), axis=1)
        self._check_input(x)

        self._train_phase_started = True
        self._train_seq[self._train_phase][0](self._refcast(x),
                                              n_updates=n_updates,
                                              epsilon=epsilon,
                                              decay=decay,
                                              momentum=momentum,
                                              verbose=verbose)



from builtins import zip
from past.builtins import basestring
from builtins import object
__docformat__ = "restructuredtext en"

from shogun import (Kernel as sgKernel,
                    Features as sgFeatures,
                    Classifier as sgClassifier)

import mdp
from mdp.utils import OrderedDict as _OrderedDict

from .svm_classifiers import _SVMClassifier, _LabelNormalizer

# switch off spurious warnings from shogun
import warnings
warnings.filterwarnings('ignore',
                        '.*Perceptron algorithm did not converge after.*',
                        RuntimeWarning)

# maybe integrate to the class
def is_shogun_classifier(test_classifier):
    """Check, if a class is a subclass of a SHOGUN classifier."""
    try:
        return issubclass(test_classifier, sgClassifier.Machine)
    except (TypeError, NameError):
        # need to fetch NameError for some swig reasons
        return False

default_shogun_classifiers = []
for labels in dir(sgClassifier):
    test_classifier = getattr(sgClassifier, labels)
    if is_shogun_classifier(test_classifier):
        default_shogun_classifiers.append(test_classifier)

shogun_classifier_types = {}
for ct in dir(sgClassifier):
    if ct.startswith("CT_"):
        shogun_classifier_types[getattr(sgClassifier, ct)] = ct


class Classifier(object):
    """This Classifier class hides the logic of setting a classifier and kernel
    from the main classifier node.
    It is currently not intended to be used on its own and might, in fact, get
    completely integrated into the ShogunSVMClassifier class."""

    def __init__(self):
        self._class = None
        self._instance = None

    def set_classifier(self, classifier, args=None):
        """Sets and initialises the classifier. If a classifier is reset
        by the user, the parameters will have to be set again.
        'classifier' can be a string, a subclass of shogun.Classifier
        or an instance of such a class.
        Note that some classifiers require initialisation arguments
        while others don't. Also note that neither labels nor feature
        vectors should be given with 'args'.

        classifier -- the SHOGUN classifier to use
        args       -- the list of arguments needed for the SHOGUN
                      constructor method of the classifier
        """
        if args is None:
            args = []
        self._class = None
        self._instance = None

        # If name is a classifier instance: Take it
        if isinstance(classifier, sgClassifier.Machine):
            self._class = classifier.__class__
            self._instance = classifier

        # If name is a class: Make an instance
        elif classifier.__class__ == type:
            if is_shogun_classifier(classifier):
                try:
                    self._class = classifier
                    self._instance = self._class(*args)
                except AttributeError:
                    msg = "Library '%s' could not be instantiated. Abstract class?" % classifier
                    raise mdp.NodeException(msg)
            else:
                msg = "The classifier '%s' is no subclass of Machine." % self._class.__name__
                raise mdp.NodeException(msg)

        # If classifier is a string: Check, if it's the name of a default library
        elif isinstance(classifier, basestring):
            possibleClasses = [labels for labels in default_shogun_classifiers
                               if labels.__name__.lower() == classifier.lower()]

            if not len(possibleClasses):
                msg = "Library '%s' is not a known subclass of Machine." % classifier
                raise mdp.NodeException(msg)

            # Take the first classifier which works
            for labels in possibleClasses:
                try:
                    self._instance = labels(*args)
                    self._class = labels
                except AttributeError:
                    # we might have a virtual class here
                    pass
            if not self._instance:
                msg = "Library '%s' could not be instantiated. Abstract class?" % classifier
                raise mdp.NodeException(msg)

        if not self._class or not self._instance:
            msg = "The classifier '%s' is not supported." % classifier
            raise mdp.NodeException(msg)

        if self.classifier_type() == "CT_NONE":
            msg = "The classifier '%s' is not valid." % classifier
            raise mdp.NodeException(msg)


    def classifier_type(self):
        """Returns the SHOGUN classifier type as a string."""
        return shogun_classifier_types[self._instance.get_classifier_type()]

    def set_param(self, param, *value):
        """Sets parameters for the classifier.

        This calls set_param(*value) on the classifier instance.
        """
        # Non-standard cases
        if param == "C" and len(value) == 1:
            # set_C needs two arguments, but if we get only one,
            # we call set_C(arg, arg)
            value += value
        # get the parameter setting method
        meth = getattr(self._instance, "set_" + param)
        # call it
        meth(*value)

    def get_param(self, param, *args):
        """Returns the parameter for a with a given name.

        This calls get_param(*args) on the classifier instance.

        """
        meth = getattr(self._instance, "get_" + param)
        return meth(*args)

    def set_train_features(self, features, labels):
        if issubclass(self._class, sgClassifier.LinearMachine):
            self._instance.set_features(features)
        elif issubclass(self._class, sgClassifier.KernelMachine):
            try:
                self.kernel.init(features, features)
            except AttributeError:
                msg = "You must assign a kernel before training."
                raise mdp.NodeException(msg)
            #self.svm.set_kernel(self.kernel)
        else:
            msg = "Sorry, SHOGUN classifiers of this type are not yet implemented"
            raise mdp.NodeException(msg)

        self._instance.set_labels(labels)

    def train(self):
        self._instance.train()

    def label(self, test_features):
        #return self._instance.classify(test_features).get_labels()
        return self._instance.apply(test_features).get_labels()

    @property
    def takes_kernel(self):
        """Returns true, if the current classifier is a kernel machine."""
        return issubclass(self._class, sgClassifier.KernelMachine)

    def _get_kernel(self):
        """Retrieve the currently set kernel from the classifier instance."""
        try:
            return self._instance.get_kernel()
        except AttributeError:
            msg = "Error retrieving kernel. The classifier might not support kernels."
            raise mdp.NodeException(msg)

    def _set_kernel(self, kernel):
        """Set the kernel in the current classifier instance"""
        try:
            self._instance.set_kernel(kernel)
        except AttributeError:
            msg = "Error setting kernel. The classifier might not support kernels."
            raise mdp.NodeException(msg)

    kernel = property(_get_kernel, _set_kernel)

    def __str__(self):
        str = self._instance.__class__.__name__
        try:
            str = str + " with %s kernel" % self.kernel.get_name()
        except (mdp.NodeException, AttributeError):
            pass
        return str


class ShogunSVMClassifier(_SVMClassifier):
    """The ``ShogunSVMClassifier`` works as a wrapper class for accessing
    the SHOGUN machine learning toolbox for support vector machines.

    Most kernel machines and linear classifier should work with this class.

    Currently, distance machines such as the K-means classifier
    are not supported yet.

    Information to paramters and additional options can be found on
    http://www.shogun-toolbox.org/

    Note that some parts in this classifier might receive some
    refinement in the future.

    This node depends on ``shogun``.
    """

    default_parameters = {
        'C': 1,
        'epsilon': 1e-3,
    }

    # Swig-code does not work with named parameters, so we have to define an order
    kernel_parameters = {
        # Simple float64t kernels
        'Chi2Kernel': [('size', 10), ('width', 1.4)],
        'GaussianKernel': [('size', 10), ('width', 1.9)],
        'LinearKernel': [],
        'PolyKernel': [('size', 10), ('degree', 3), ('inhomogene', True)],
        'PyramidChi2': [('size',), ('num_cells2',),
                        ('weights_foreach_cell2',), ('width_computation_type2',),
                        ('width2',)],
        'SigmoidKernel': [('size', 10), ('gamma', 1), ('coef0', 0)]
    }

    def __init__(self, classifier="libsvmmulticlass", classifier_arguments=(),
                 classifier_options=None, kernel_name=None, kernel_options=None,
                 num_threads="autodetect",
                 input_dim=None, output_dim=None, dtype=None):
        """
        Initialises a new ShogunSVMClassifier.

        :Arguments:
          classifier
            The classifier to use
          classifier_arguments
            Arguments needed for the constructor of the classifier
          classifier_options
            Options for the classifier
          num_threads
            The number of threads, SHOGUN should use
            can be set to ``"autodetect"``, then SHOGUN will use
            the number of CPU cores.

            Attention: this could crash on windows
        """
        super(ShogunSVMClassifier, self).__init__(input_dim=input_dim,
                                                  output_dim=output_dim,
                                                  dtype=dtype)

        if classifier_options is None:
            classifier_options = {}

        self.classifier = Classifier()
        self.classifier.set_classifier(classifier, classifier_arguments)
        self.classifier_options = self.default_parameters
        self.classifier_options.update(classifier_options)

        for p in list(self.classifier_options.keys()):
            try:
                self.set_classifier_param(p, self.classifier_options[p])
            except Exception:
                pass

        self._num_threads = num_threads
        self._set_num_threads()

        if kernel_name:
            self.set_kernel(kernel_name, kernel_options)

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        # Support only float64 because of external library
        return ('float64',)

    def _set_num_threads(self):
        # init number of threads
        if self._num_threads == "autodetect":
            try:
                self._num_threads = self.classifier._instance.parallel.get_num_cpus()
            except SystemError:
                # We're helping shogun here
                self._num_threads = 1
        self.classifier._instance.parallel.set_num_threads(self._num_threads)

    def set_classifier_param(self, param, *value):
        """Sets parameters for the classifier.
        """
        self.classifier.set_param(param, *value)

    def set_kernel(self, kernel_name, kernel_options=None):
        """Sets the Kernel along with options.
        'options' must be a tuple with the arguments of the kernel constructor
        in SHOGUN.
        We try to guess it right in many cases but in general, you will have to
        consult the SHOGUN documentation.

        :Parameters:
          kernel_name
            The kernel to use. Default parameters are specified for

            - ``"PolyKernel"``
            - ``"GaussianKernel"``
            - ``"LinearKernel"``
            - ``"SigmoidKernel"``

            Further kernels are possible if they are included in
            SHOGUN and if kernel_options provides the correct init
            arguments.

          kernel_options
            For known kernels, a dict specifying the options is
            possible. Options not included take a default value.
            Unknown kernels need an ordered list of constructor
            arguments.
        """
        if kernel_options is None:
            kernel_options = {}
        if kernel_name in ShogunSVMClassifier.kernel_parameters \
            and not isinstance(kernel_options, list):
            default_opts = _OrderedDict(ShogunSVMClassifier.kernel_parameters[kernel_name])
            default_opts.update(kernel_options)
            options = list(default_opts.values())

        kernel_meth = getattr(sgKernel, kernel_name)
        try:
            kernel = kernel_meth(*options)
        except NotImplementedError as msg:
            msg = ("Tried to call %s with arguments %s\n" %
                   (kernel_meth.__module__ + '.' + kernel_meth.__name__,
                    tuple(options).__repr__()) +
                   "Got the following error message:\n" + msg.__str__())
            raise mdp.NodeException(msg)
        self.classifier.kernel = kernel

    def _stop_training(self):
        super(ShogunSVMClassifier, self)._stop_training()
        self.normalizer = _LabelNormalizer(self.labels)
        labels = self.normalizer.normalize(self.labels)
        # shogun expects float labels
        labels = sgFeatures.Labels(labels.astype(float))

        features = sgFeatures.RealFeatures(self.data.transpose())

        self.classifier.set_train_features(features, labels)
        self.classifier.train()

    def training_set(self, ordered=False):
        """Shows the set of data that has been inserted to be trained."""
        if ordered:
            labels = set(self.labels)
            data = {}
            for l in labels:
                data[l] = []
            for k, v in zip(self.labels, self.data):
                data[k].append(v)
            return data
        else:
            return list(zip(self.labels, self.data))

    def _label(self, x):
        """Classify the input data 'x'
        """
        test_features = sgFeatures.RealFeatures(x.transpose())

        labels = self.classifier.label(test_features)

        if self.normalizer:
            return self.normalizer.revert(labels)
        else:
            return labels



import mdp
import os
from subprocess import Popen, PIPE, STDOUT

def get_git_revision():
    """When mdp is run from inside a git repository, this function
    returns the current revision that git-describe gives us.

    If mdp is installed (or git fails for some other reason),
    an empty string is returned.
    """
    # TODO: Introduce some fallback method that takes the info from a file
    revision = ''
    try:
        # we need to be sure that we call from the mdp dir
        mdp_dir = os.path.dirname(mdp.__file__)
        # --tags ensures that most revisions have a name even without
        # annotated tags
        # --dirty=+ appends a plus if the working copy is modified
        command = ["git", "describe", "--tags", "--dirty=+"]
        proc = Popen(command, stdout=PIPE, stderr=STDOUT, cwd=mdp_dir,
                     universal_newlines=True)
        exit_status = proc.wait()
        # only get the revision if command succeded
        if exit_status == 0:
            revision = proc.stdout.read().strip()
    except OSError:
        pass
    return revision
